from lxml import etree


def format_xml(xml_string: str, minify: bool = False) -> str:
    """
    Formats XML string, making it pretty or minified
    """

    # Parse string & remove blank lines
    root = etree.XML(xml_string, parser = etree.XMLParser(remove_blank_text = True))

    # Format XML string
    return etree.tostring(root, pretty_print = not minify, encoding = 'unicode')

    # For more information,
    # see http://www.ronrothman.com/public/leftbraned/xml-dom-minidom-toprettyxml-and-silly-whitespace



def extract_data(args: list) -> dict:
    """
    Processes unknown `click` arguments
    """

    # Create data array
    data = {}

    # Iterate over unknown arguments and ..
    for i, arg in enumerate(args):
        # (1) .. skip values & shorthand arguments
        if arg[:2] != '--':
            continue

        # (2) .. attempt to ..
        try:
            # .. treat them as options
            if args[i + 1][:2] != '--':
                data[arg[2:]] = args[i + 1]

            # .. unless they are flags
            else:
                data[arg[2:]] = True

        # .. otherwise ..
        except IndexError:
            # .. treat them as flags
            data[arg[2:]] = True

    return data

# For more information,
# see https://stackoverflow.com/a/32946412


def font_metrics(font_path: str, font_size: int) -> dict:
    """
    Extracts font metrics
    """

    # Import modules
    from fontline.metrics import MetricsObject
    from fontTools import ttLib

    # Gather information about given font
    # (1) Initialize tooling
    tt = ttLib.TTFont(font_path)
    metrics = MetricsObject(tt, font_path)

    # (2) Extract relevant metrics
    return {
        'metrics': metrics,
        'height': font_size * metrics.hheaascdesc_to_upm,
        'x_height': font_size * metrics.os2_x_height / metrics.units_per_em,
        'ascent': font_size * metrics.hhea_ascent / metrics.units_per_em,
        'descent': font_size * abs(metrics.hhea_descent) / metrics.units_per_em,
    }



from os.path import abspath, basename, dirname, isdir, isfile, join
from copy import deepcopy
from glob import glob
from json import load as load_json

from click import get_app_dir
from diskcache import Cache
from jinja2 import Template
from requests import get

from .utils import font_metrics
from .utils.svg import text2path
from .utils.xml import format_xml


class Badges():
    # Built-in fonts
    fonts = {
        # (1) Hack
        # See https://github.com/source-foundry/Hack
        'hack-regular': 'hack/Hack-Regular.ttf',
        'hack-bold': 'hack/Hack-Bold.ttf',
        'hack-italic': 'hack/Hack-Italic.ttf',
        'hack-bolditalic': 'hack/Hack-BoldItalic.ttf',

        # (2) Inter
        # See https://github.com/rsms/inter
        'inter-thin': 'inter/Inter-Thin.otf',
        'inter-thinitalic': 'inter/Inter-ThinItalic.otf',
        'inter-extralight': 'inter/Inter-ExtraLight.otf',
        'inter-extralightitalic': 'inter/Inter-ExtraLightItalic.otf',
        'inter-light': 'inter/Inter-Light.otf',
        'inter-lightitalic': 'inter/Inter-LightItalic.otf',
        'inter-regular': 'inter/Inter-Regular.otf',
        'inter-italic': 'inter/Inter-Italic.otf',
        'inter-medium': 'inter/Inter-Medium.otf',
        'inter-mediumitalic': 'inter/Inter-MediumItalic.otf',
        'inter-semibold': 'inter/Inter-SemiBold.otf',
        'inter-semibolditalic': 'inter/Inter-SemiBoldItalic.otf',
        'inter-bold': 'inter/Inter-Bold.otf',
        'inter-bolditalic': 'inter/Inter-BoldItalic.otf',
        'inter-extrabold': 'inter/Inter-ExtraBold.otf',
        'inter-extrabolditalic': 'inter/Inter-ExtraBoldItalic.otf',
        'inter-black': 'inter/Inter-Black.otf',
        'inter-blackitalic': 'inter/Inter-BlackItalic.otf',
    }


    # Cache expiry time (in seconds)
    ttl = 60 * 60 * 24


    def __init__(self, template: str = 'standard', cache_dir: str = None) -> None:
        # Load template configuration
        self.template = template

        # Initialize file cache
        self.cache = Cache(cache_dir or get_app_dir('py-badges'))


    def get_template_dir(self, template: str) -> str:
        # Set base directory
        base_dir = join(abspath(dirname(__file__)), 'data', 'templates')

        # Assume 'template' representing directory
        template_dir = template

        # If built-in template is available ..
        if template and isdir(join(base_dir, template)):
            # .. load it
            template_dir = join(base_dir, template)

        # If directory does not exist ..
        if not template_dir or not isdir(template_dir):
            # .. raise exception
            raise ValueError('Cannot load template: "%s"' % template)

        return template_dir


    def get_theme(self, theme: str) -> str:
        # Assume 'theme' representing file path
        theme_file = theme

        # If theme is supported ..
        if theme in self.themes:
            # .. load it
            theme_file = join(self.template_dir, 'themes', '%s.json' % theme)

        # If theme file does not exist ..
        if not isfile(theme_file):
            # .. raise exception
            raise ValueError('Cannot load theme: "%s"' % theme)

        return theme_file


    def get_font(self, font: str) -> str:
        # Set base directory
        base_dir = join(abspath(dirname(__file__)), 'data', 'fonts')

        # Assume 'font' representing file path
        font_file = font

        # If font is supported ..
        if font.lower() in self.fonts:
            # .. load it
            font_file = join(base_dir, self.fonts[font.lower()])

        # If font file does not exist ..
        if not isfile(font_file):
            # .. raise exception
            raise ValueError('Cannot load font: "%s"' % font)

        return font_file


    @property
    def template(self) -> str:
        return basename(self.template_dir)


    @template.setter
    def template(self, template: str) -> None:
        # Determine template directory
        template_dir = self.get_template_dir(template)

        # Check if required template files exist ..
        for file in ['config.json', 'template.j2']:
            if not isfile(join(template_dir, file)):
                raise ValueError('Missing required template file: "%s"' % file)

        # Process template files
        # (1) Load configuration file
        with open(join(template_dir, 'config.json'), 'r') as file:
            self.config = load_json(file)

        # (2) Load jinja template
        with open(join(template_dir, 'template.j2'), 'r') as file:
            self.jinja_template = Template(file.read())

            # Provide template functions
            self.jinja_template.globals.update(
                int = int,
                str = str,
                float = float,
                text2path = text2path,
                font_metrics = self.font_metrics,
                render_text = self.render_text
            )

        # Reset themes
        self.themes = []

        # (3) Load themes
        # If theme directory exists ..
        if isdir(join(template_dir, 'themes')):
            # .. assign available themes
            self.themes = [basename(theme)[:-5] for theme in glob(join(template_dir, 'themes', '*.json'))]

        # Assign template directory
        self.template_dir = template_dir


    def render_text(self, text: str, font: str, font_size: int, x: int = 0, y: int = 0) -> dict:
        # Import modules
        from svgpathtools.parser import parse_path

        # Create path
        path = text2path(text, self.get_font(font), font_size, x, y)

        # Determine width & height of path
        x_min, x_max, y_min, y_max = parse_path(path).bbox()

        return {
            'path': '<path d="%s" />' % path,
            'width': x_max - x_min,
            'height': y_max - y_min,
        }


    def font_metrics(self, font: str, font_size: int) -> dict:
        # Determine font file & provide metrics
        return font_metrics(self.get_font(font), font_size)


    def fetch(self, url: str, content_type: str = 'json') -> dict:
        # Check if cache entry exists ..
        if url not in self.cache:
            # ..otherwise create it
            # (1) Fetch data from URL
            response = get(url)

            # (2) Validate it
            if response.status_code != 200:
                raise Exception('Access to API failed with error code: "%s"' % response.reason)

            # (3) Respect content type
            response = response.json() if content_type == 'json' else response.text

            # (4) Create cache entry for URL
            self.cache.set(url, response, self.ttl)

        return self.cache[url]


    def render(self, data: dict, theme: str = None, provider: str = None, minify: bool = False) -> str:
        # Load template configuration
        config = deepcopy(self.config)

        # If no theme provided but template specifies default ..
        if not theme and 'theme' in config:
            # .. fallback to its default theme
            theme = config['theme']

        # Let template know current theme being used
        config['theme'] = theme

        # If theme provided ..
        if theme:
            # .. update configuration
            with open(self.get_theme(theme), 'r') as file:
                config.update(load_json(file))

        # Include custom settings like data or theme options
        config.update(data)

        # If (valid) provider provided ..
        if provider:
            # (1) .. check for repository ..
            if not 'repo' in data:
                # .. otherwise throw exception
                raise Exception('No "repo" specified for provider: "%s"' % provider)

            # (2) .. create data array
            repo = None

            # (3) .. fetch data for current repository
            if provider.lower() in ['cb', 'codeberg']:
                # Popular placeholders:
                #
                # {size}
                # {stars_count}
                # {forks_count}
                # {watchers_count}
                # {open_issues_count}
                # {open_pr_counter}
                # {release_counter}
                # {default_branch}
                repo = self.fetch('https://codeberg.org/api/v1/repos/%s' % data['repo'])

                # For more information,
                # see https://try.gitea.io/api/swagger

                # Add status for 'Woodpecker CI'
                #
                # Placeholder:
                #
                # {status}
                status = 'none'

                svg_string = self.fetch('https://ci.codeberg.org/api/badges/%s/status.svg' % data['repo'], 'text')

                if 'success' in svg_string:
                    status = 'success'

                if 'failure' in svg_string:
                    status = 'failure'

                # Store status information
                repo['status'] = status

            if provider.lower() in ['gh', 'github']:
                # Popular placeholders:
                #
                # {size}
                # {stargazers_count}
                # {watchers_count}
                # {language}
                # {forks_count}
                # {open_issues_count}
                # {default_branch}
                # {license/key}
                # {license/name}
                # {license/spdx_id}
                repo = self.fetch('https://api.github.com/repos/%s' % data['repo'])

                # For more information,
                # see https://docs.github.com/en/rest/guides/getting-started-with-the-rest-api

            # If no repository data is available ..
            if not repo or ('message' in repo and repo['message'].lower() == 'not found'):
                # .. abort execution
                raise Exception('No repository data available: "%s"' % data['repo'])

            # Iterate over data items but skip ..
            for key, value in config.items():
                # (1) .. non-strings
                if not isinstance(value, str):
                    continue

                # (2) .. non-placeholders
                if value[:1] != '{':
                    continue

                # (3) .. reserved keywords
                if key in ['repo']:
                    continue

                # Format placeholder value
                value = value.strip('{}')

                # If placeholder value is present ..
                if value in repo:
                    # .. replace matching entries with repository data
                    config[key] = str(repo[value])

                # Look for nested values
                if '/' in value:
                    value = value.split('/')

                    # If nested placeholder value is present ..
                    if value[0] in repo and value[1] in repo[value[0]]:
                        # .. replace matching entries with nested repository data
                        config[key] = repo[value[0]][value[1]]

        # Render SVG string using data & theme settings
        return format_xml(self.jinja_template.render(data = config), minify)



import click

from . import Badges
from .utils import extract_data


@click.command(context_settings = dict(ignore_unknown_options = True,allow_extra_args = True))
@click.argument('output', type = click.File('w'))
@click.option('-s', '--style', help = 'Set style (name or directory).')
@click.option('-t', '--theme', help = 'Set theme (name or file path).')
@click.option('-p', '--provider', help = 'Set provider (Codeberg, GitHub, ..).')
@click.option('-m', '--minify', is_flag = True, help = 'Whether to minify SVG output.')
@click.option('-v', '--verbose', count = True, help = 'Enable verbose mode.')
@click.version_option('0.3.0')
@click.pass_context
def cli(ctx,
    output: click.File,
    style: str,
    theme: str,
    provider: str,
    minify: bool,
    verbose: int
) -> None:
    """
    Writes SVG file to OUTPUT
    """

    # Attempt to ..
    try:
        # .. load template
        template = style or 'standard'

        # .. initalize object
        badges = Badges(template)

        # .. generate data from unknown arguments
        data = extract_data(ctx.args)

        # .. write SVG string to file
        output.write(badges.render(data, theme, provider, minify))

    # .. otherwise ..
    except Exception as error:
        # (1) .. print error message
        click.echo(error)



import setuptools

# Load README
with open('README.md', 'r', encoding = 'utf8') as file:
    long_description = file.read()

# Define package metadata
setuptools.setup(
    name = 'cb_badges',
    version = '0.3.0',
    author = 'Martin Folkers',
    author_email = 'hello@twobrain.io',
    description = '',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://codeberg.org/S1SYPHOS/cb-badges',
    license = 'MIT',
    project_urls = {
        'Issues': 'https://codeberg.org/S1SYPHOS/cb-badges/issues',
    },
    entry_points = """
        [console_scripts]
        badges=cb_badges.cli:cli
    """,
    classifiers = [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    packages = setuptools.find_packages(),
    package_data = {'badges': [
        'data/fonts/**/*.ttf',
        'data/fonts/**/*.otf',
        'data/templates/**/*.j2',
        'data/templates/**/*.json',
        'data/templates/**/themes/*.json',
    ]},
    install_requires = [
        'bs4',
        'click',
        'diskcache',
        'font-line',
        'jinja2',
        'lxml',
        'pycairo',
        'pytest',
        'requests',
        'svgpathtools',
    ],
    python_requires = '>= 3.7'
)



from io import BytesIO
from os.path import isfile
from bs4 import BeautifulSoup as bs
from cairo import Context, FONT_SLANT_NORMAL, FONT_WEIGHT_NORMAL, SVGSurface

from .cairo import create_cairo_font_face_for_file


def text2path(
    text: str,
    font: str,
    font_size: int = 16,
    x: int = 0,
    y: int = 0,
    width: int = 1000,
    height: int = 1000
) -> str:
    """
    Converts text to SVG path
    """

    # Create file-like SVG object ..
    pathio = BytesIO()

    # .. from `text`
    with SVGSurface(pathio, width, height) as surface:
        # Create drawing
        ctx = Context(surface)

        # If font file was provided ..
        if isfile(font):
            # .. load it
            ctx.set_font_face(create_cairo_font_face_for_file(font))

        # .. otherwise ..
        else:
            # .. treat it as font title
            ctx.select_font_face(font, FONT_SLANT_NORMAL, FONT_WEIGHT_NORMAL)

        # Set font size
        ctx.set_font_size(font_size)

        # Fix `pycairo` being (sometimes) unable to convert (certain) integers
        if text.isdigit() and y == 0:
            y = 1

        # Get ready ..
        ctx.move_to(x, y)

        # .. DRAW!
        ctx.text_path(text)

        # Fill it
        ctx.fill()

    # Set pointer to beginning
    pathio.seek(0)

    # Fetch first path element ..
    path = bs(pathio.read(), 'lxml').find('path')

    # .. but fail if none present
    if not path:
        raise Exception('Cannot render text: "%s"' % text)

    # Extract path defintion
    return path['d']
